# Iter\[A: A\]

Wrapper class containing methods to modify iterators.


```pony
class ref Iter[A: A] is
  Iterator[A] ref
```

#### Implements

* [Iterator](builtin-Iterator)\[A\] ref

---

## Constructors

### create

```pony
new ref create(
  iter: Iterator[A] ref)
: Iter[A] ref^
```
#### Parameters

*   iter: [Iterator](builtin-Iterator)\[A\] ref

#### Returns

* [Iter](itertools-Iter)\[A\] ref^

---

### chain

Take an iterator of iterators and work through them, returning the
items of the first one, then the second one, and so on.

## Example
```pony
let xs = [as I64: 1, 2].values()
let ys = [as I64: 3, 4].values()

Iter[I64].chain([xs, ys].values())
```
`1 2 3 4`


```pony
new ref chain(
  outer_iterator: Iterator[Iterator[A] ref] ref)
: Iter[A] ref^
```
#### Parameters

*   outer_iterator: [Iterator](builtin-Iterator)\[[Iterator](builtin-Iterator)\[A\] ref\] ref

#### Returns

* [Iter](itertools-Iter)\[A\] ref^

---

## Public Functions

### has_next

```pony
fun ref has_next()
: Bool val
```

#### Returns

* [Bool](builtin-Bool) val

---

### next

```pony
fun ref next()
: A ?
```

#### Returns

* A ?

---

### all

Return false if at least one value of the iterator fails to match the
predicate `f`. This method short-circuits at the first value where the
predicate returns false, otherwise true is returned.

## Examples

```pony
Iter[I64]([as I64: 2, 4, 6].values())
  .all({(x: I64): Bool => (x % 2) == 0 })
```
`true`

```pony
Iter[I64]([as I64: 2, 3, 4].values())
  .all({(x: I64): Bool => (x % 2) == 0 })
```
`false`


```pony
fun ref all(
  f: {(A!): Bool ?}[A] box)
: Bool val
```
#### Parameters

*   f: {(A!): Bool ?}[A] box

#### Returns

* [Bool](builtin-Bool) val

---

### any

Return true if at least one value of the iterator matches the predicate
`f`. This method short-circuits at the first value where the predicate
returns true, otherwise false is returned.

## Examples

```pony
Iter[I64]([as I64: 2, 4, 6].values())
  .any({(x: I64): Bool => (x % 2) == 1 })
```
`false`

```pony
Iter[I64]([as I64: 2, 3, 4].values())
  .any({(x: I64): Bool => (x % 2) == 1 })
```
`true`


```pony
fun ref any(
  f: {(A!): Bool ?}[A] box)
: Bool val
```
#### Parameters

*   f: {(A!): Bool ?}[A] box

#### Returns

* [Bool](builtin-Bool) val

---

### collect\[optional B: [Seq](builtin-Seq)\[A!\] ref\]

Push each value from the iterator into the collection `coll`.

## Example

```pony
Iter[I64]([as I64: 1, 2, 3].values())
  .collect(Array[I64](3))
```
`[1, 2, 3]`


```pony
fun ref collect[optional B: Seq[A!] ref](
  coll: B ref)
: B ref^
```
#### Parameters

*   coll: B ref

#### Returns

* B ref^

---

### count

Return the number of values in the iterator.

## Example

```pony
Iter[I64]([as I64: 1, 2, 3].values())
  .count()
```
`3`


```pony
fun ref count()
: USize val
```

#### Returns

* [USize](builtin-USize) val

---

### cycle

Repeatedly cycle through the values from the iterator.

WARNING: The values returned by the original iterator are cached, so
the input iterator should be finite.

## Example

```pony
Iter[I64]([as I64: 1, 2, 3].values())
  .cycle()
```
`1 2 3 1 2 3 1 2 3 ...`


```pony
fun ref cycle()
: Iter[A!] ref^
```

#### Returns

* [Iter](itertools-Iter)\[A!\] ref^

---

### enum\[optional B: ([Real](builtin-Real)\[B val\] val & ([I8](builtin-I8) val | [I16](builtin-I16) val | [I32](builtin-I32) val | 
    [I64](builtin-I64) val | [I128](builtin-I128) val | [ILong](builtin-ILong) val | 
    [ISize](builtin-ISize) val | [U8](builtin-U8) val | [U16](builtin-U16) val | 
    [U32](builtin-U32) val | [U64](builtin-U64) val | [U128](builtin-U128) val | 
    [ULong](builtin-ULong) val | [USize](builtin-USize) val | [F32](builtin-F32) val | 
    [F64](builtin-F64) val))\]

An iterator which yields the current iteration count as well as the next
value from the iterator.

## Example

```pony
Iter[I64]([as I64: 1, 2, 3].values())
  .enum()
```
`(0, 1) (1, 2) (2, 3)`


```pony
fun ref enum[optional B: (Real[B val] val & (I8 val | I16 val | I32 val | 
    I64 val | I128 val | ILong val | 
    ISize val | U8 val | U16 val | 
    U32 val | U64 val | U128 val | 
    ULong val | USize val | F32 val | 
    F64 val))]()
: Iter[(B val , A)] ref^
```

#### Returns

* [Iter](itertools-Iter)\[(B val , A)\] ref^

---

### filter

Return an iterator that only returns items that match the predicate `f`.

## Example

```pony
Iter[I64]([as I64: 1, 2, 3, 4, 5, 6].values())
  .filter({(x: I64): Bool => (x % 2) == 0 })
```
`2 4 6`


```pony
fun ref filter(
  f: {(A!): Bool ?}[A] box)
: Iter[A] ref^
```
#### Parameters

*   f: {(A!): Bool ?}[A] box

#### Returns

* [Iter](itertools-Iter)\[A\] ref^

---

### find

Return the nth value in the iterator that satisfies the predicate `f`.

## Examples

```pony
Iter[I64]([as I64: 1, 2, 3].values())
  .find({(x: I64): Bool => (x % 2) == 0 })
```
`2`

```pony
Iter[I64]([as I64: 1, 2, 3, 4].values())
  .find({(x: I64): Bool => (x % 2) == 0 }, 2)
```
`4`


```pony
fun ref find(
  f: {(A!): Bool ?}[A] box,
  n: USize val = 1)
: A ?
```
#### Parameters

*   f: {(A!): Bool ?}[A] box
*   n: [USize](builtin-USize) val = 1

#### Returns

* A ?

---

### fold\[B: B\]

Apply a function to every element, producing an accumulated value.

## Example

```pony
Iter[I64]([as I64: 1, 2, 3].values())
  .fold[I64]({(sum: I64, x: I64): I64 => sum + x }, 0)
```
`6`


```pony
fun ref fold[B: B](
  f: {(B, A!): B^ ?}[A, B] box,
  acc: B)
: B^ ?
```
#### Parameters

*   f: {(B, A!): B^ ?}[A, B] box
*   acc: B

#### Returns

* B^ ?

---

### last

Return the last value of the iterator.

## Example

```pony
Iter[I64]([as I64: 1, 2, 3].values())
  .last()
```
`3`


```pony
fun ref last()
: A ?
```

#### Returns

* A ?

---

### map\[B: B\]

Return an iterator where each item's value is the application of the given
function to the value in the original iterator.

## Example

```pony
Iter[I64]([as I64: 1, 2, 3].values())
  .map[I64]({(x: I64): I64 => x * x })
```
`1 4 9`


```pony
fun ref map[B: B](
  f: {(A!): B ?}[A, B] box)
: Iter[B] ref^
```
#### Parameters

*   f: {(A!): B ?}[A, B] box

#### Returns

* [Iter](itertools-Iter)\[B\] ref^

---

### nth

Return the nth value of the iterator.

## Example

```pony
Iter[I64]([as I64: 1, 2, 3].values())
  .nth(2)
```
`2`


```pony
fun ref nth(
  n: USize val)
: A ?
```
#### Parameters

*   n: [USize](builtin-USize) val

#### Returns

* A ?

---

### run

Iterate through the values of the iterator without a for loop. The
function `on_error` will be called if the iterator's `has_next` method
returns true but its `next` method trows an error.

## Example

```pony
Iter[I64]([as I64: 1, 2, 3].values())
  .map[None]({(x: I64) => env.out.print(x.string()) })
  .run()
```
```
1
2
3
```


```pony
fun ref run(
  on_error: {()}[A] box = lambda)
: None val
```
#### Parameters

*   on_error: {()}[A] box = lambda

#### Returns

* [None](builtin-None) val

---

### skip

Skip the first n values of the iterator.

## Example

```pony
Iter[I64]([as I64: 1, 2, 3, 4, 5, 6].values())
  .skip(3)
```
`4 5 6`


```pony
fun ref skip(
  n: USize val)
: Iter[A] ref^
```
#### Parameters

*   n: [USize](builtin-USize) val

#### Returns

* [Iter](itertools-Iter)\[A\] ref^

---

### skip_while

Skip values of the iterator while the predicate `f` returns true.

## Example

```pony
Iter[I64]([as I64: 1, 2, 3, 4, 5, 6].values())
  .skip_while({(x: I64): Bool => x < 4 })
```
`4 5 6`


```pony
fun ref skip_while(
  f: {(A!): Bool ?}[A] box)
: Iter[A] ref^
```
#### Parameters

*   f: {(A!): Bool ?}[A] box

#### Returns

* [Iter](itertools-Iter)\[A\] ref^

---

### take

Return an iterator for the first n elements.

## Example

```pony
Iter[I64]([as I64: 1, 2, 3, 4, 5, 6].values())
  .take(3)
```
`1 2 3`


```pony
fun ref take(
  n: USize val)
: Iter[A] ref^
```
#### Parameters

*   n: [USize](builtin-USize) val

#### Returns

* [Iter](itertools-Iter)\[A\] ref^

---

### take_while

Return an iterator that returns values while the predicate `f` returns
true.

## Example

```pony
Iter[I64]([as I64: 1, 2, 3, 4, 5, 6].values())
  .take_while({(x: I64): Bool => x < 4 })
```
`1 2 3`


```pony
fun ref take_while(
  f: {(A!): Bool ?}[A] box)
: Iter[A] ref^
```
#### Parameters

*   f: {(A!): Bool ?}[A] box

#### Returns

* [Iter](itertools-Iter)\[A\] ref^

---

### zip\[B: B\]

Zip two iterators together so that each call to next() results in the
a tuple with the next value of the first iterator and the next value
of the second iterator. The number of items returned is the minimum of
the number of items returned by the two iterators.

## Example

```pony
Iter[I64]([as I64: 1, 2].values())
  .zip[I64]([as I64: 3, 4].values())
```
`(1, 3) (2, 4)`


```pony
fun ref zip[B: B](
  i2: Iterator[B] ref)
: Iter[(A , B)] ref^
```
#### Parameters

*   i2: [Iterator](builtin-Iterator)\[B\] ref

#### Returns

* [Iter](itertools-Iter)\[(A , B)\] ref^

---

### zip2\[B: B, C: C\]

Zip three iterators together so that each call to next() results in
the a tuple with the next value of the first iterator, the next value
of the second iterator, and the value of the third iterator. The
number of items returned is the minimum of the number of items
returned by the three iterators.


```pony
fun ref zip2[B: B, C: C](
  i2: Iterator[B] ref,
  i3: Iterator[C] ref)
: Iter[(A , B , C)] ref^
```
#### Parameters

*   i2: [Iterator](builtin-Iterator)\[B\] ref
*   i3: [Iterator](builtin-Iterator)\[C\] ref

#### Returns

* [Iter](itertools-Iter)\[(A , B , C)\] ref^

---

### zip3\[B: B, C: C, D: D\]

Zip four iterators together so that each call to next() results in the
a tuple with the next value of each of the iterators. The number of
items returned is the minimum of the number of items returned by the
iterators.


```pony
fun ref zip3[B: B, C: C, D: D](
  i2: Iterator[B] ref,
  i3: Iterator[C] ref,
  i4: Iterator[D] ref)
: Iter[(A , B , C , D)] ref^
```
#### Parameters

*   i2: [Iterator](builtin-Iterator)\[B\] ref
*   i3: [Iterator](builtin-Iterator)\[C\] ref
*   i4: [Iterator](builtin-Iterator)\[D\] ref

#### Returns

* [Iter](itertools-Iter)\[(A , B , C , D)\] ref^

---

### zip4\[B: B, C: C, D: D, E: E\]

Zip five iterators together so that each call to next() results in the
a tuple with the next value of each of the iterators. The number of
items returned is the minimum of the number of items returned by the
iterators.


```pony
fun ref zip4[B: B, C: C, D: D, E: E](
  i2: Iterator[B] ref,
  i3: Iterator[C] ref,
  i4: Iterator[D] ref,
  i5: Iterator[E] ref)
: Iter[(A , B , C , D , E)] ref^
```
#### Parameters

*   i2: [Iterator](builtin-Iterator)\[B\] ref
*   i3: [Iterator](builtin-Iterator)\[C\] ref
*   i4: [Iterator](builtin-Iterator)\[D\] ref
*   i5: [Iterator](builtin-Iterator)\[E\] ref

#### Returns

* [Iter](itertools-Iter)\[(A , B , C , D , E)\] ref^

---

